import json
from itertools import izip

import numpy as np
import scipy.sparse as sparse
from scipy.ndimage import generate_binary_structure

from lib5c.util.system import check_outdir


class DirectedDisjointSet(object):
    """
    Based on https://stackoverflow.com/a/3067672 but with two colors.

    The overall effect is like a directed sparse graph - DDS.add(a, b) is like
    adding an edge from a to b. a gets marked as a source, b does not (anything
    not in the set DDS.sources is assumed to be a destination). If b is in an
    existing group, but isn't also the source of any other edge, then the groups
    won't be merged. Finally, RBDS.get_groups() will be filtered to include only
    source nodes.

    This is an improved version where destination nodes are not stored anywhere
    if they haven't previously been seen as a source.
    """
    def __init__(self):
        self.leader = {}  # maps a member to the group's leader
        self.group = {}  # maps a group leader to the group (which is a set)

    def add(self, a, b):
        leadera = self.leader.get(a)
        leaderb = self.leader.get(b)
        if leadera is not None:
            if leaderb is not None:
                # we've seen both members before, need to merge the groups
                if leadera == leaderb:
                    return  # nothing to do
                groupa = self.group[leadera]
                groupb = self.group[leaderb]
                if len(groupa) < len(groupb):
                    a, leadera, groupa, b, leaderb, groupb = \
                        b, leaderb, groupb, a, leadera, groupa
                groupa |= groupb
                del self.group[leaderb]
                for k in groupb:
                    self.leader[k] = leadera
            else:
                # a has been seen before but b doesn't exist - do nothing
                return
        else:
            if leaderb is not None:
                # a hasn't been seen before but b exists - connect them
                self.group[leaderb].add(a)
                self.leader[a] = leaderb
            else:
                # a hasn't been seen before and b doesn't exist
                # create a new group for a, ignore b
                self.leader[a] = a
                self.group[a] = {a}

    def get_groups(self):
        return self.group.values()


def find_clusters(sig_points, connectivity=1):
    """
    Finds clusters of adjacent True points in a boolean matrix.

    Parameters
    ----------
    sig_points : scipy.sparse.spmatrix or np.ndarray
        A boolean matrix indicating which points are significant.
    connectivity : int
        The connectivity to use when clustering.

    Returns
    -------
    list of set of tuple of int
        The clusters.
    """
    spmatrix = sparse.coo_matrix(sig_points)
    structure = generate_binary_structure(2, connectivity)
    xs, ys = np.where(structure)
    xs -= 1
    ys -= 1
    shifts = [np.array(c) for c in zip(xs, ys)]
    dds = DirectedDisjointSet()
    for c in izip(spmatrix.row, spmatrix.col):
        center = np.array(c)
        for shift in shifts:
            dds.add(tuple(center), tuple(center + shift))
    return dds.get_groups()


class NumpyEncoder(json.JSONEncoder):
    """
    Pass this to `json.dump()` to correctly serialize numpy values.

    Credit: https://stackoverflow.com/a/27050186
    """
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        else:
            return super(NumpyEncoder, self).default(obj)


def save_clusters(clusters, outfile):
    """
    Saves cluster information to disk in sparse JSON format.

    Parameters
    ----------
    clusters : np.ndarray or list of set of tuple of int
        If an `np.ndarray` is passed, it should be square and triangular and
        have int dtype. Entries should be the cluster id for points which belong
        to that cluster, zero everywhere else. If a list of sets is passed, the
        sets are clusters, the tuples are the indices of entries in that
        cluster.
    outfile : str
        File to write JSON output to.
    """
    check_outdir(outfile)
    with open(outfile, 'w') as handle:
        if type(clusters) == np.ndarray:
            clusters = convert_cluster_array_to_sparse(clusters)
        json.dump([[[i, j] for i, j in c] for c in clusters], handle,
                  cls=NumpyEncoder)


def convert_cluster_array_to_sparse(cluster_array):
    """
    Converts an array of cluster information to a sparse, JSON-friendly format.

    Parameters
    ----------
    cluster_array : np.ndarray or scipy.sparse.spmatrix
        Square, triangular, int dtype. Entries should be the cluster id for
        points which belong to that cluster, zero everywhere else.

    Returns
    -------
    list of sets of tuples of int
        The sets are clusters, tuples are the matrix indices of the pixels in
        that cluster.

    Notes
    -----
    Since the introduction of `hiclite.util.clusters.find_clusters()`, this
    function is no longer used.
    """
    obj = {}
    if isinstance(cluster_array, sparse.coo_matrix):
        x = cluster_array
    elif isinstance(cluster_array, sparse.spmatrix):
        x = cluster_array.tocoo()
    else:
        x = sparse.coo_matrix(cluster_array)
    for i, j, idx in izip(x.row, x.col, x.data):
        if not idx:
            continue
        if idx not in obj:
            obj[int(idx)] = set()
        obj[int(idx)].add((int(i), int(j)))
    return obj.values()


def load_clusters(infile):
    """
    Loads clusters in a sparse format from a JSON file.

    Parameters
    ----------
    infile : str
        The JSON file containing sparse cluster information.

    Returns
    -------
    list of set of tuple of int
        The sets are clusters, the tuples are the indices of entries in that
        cluster.
    """
    with open(infile, 'r') as handle:
        return [set([tuple(e) for e in cluster])
                for cluster in json.load(handle)]


def clusters_to_coo(clusters, shape):
    """
    Converts clusters (list of list of tuple) to a COO sparse matrix.

    Parameters
    ----------
    clusters : list of list of tuple
        The outer list is a list of clusters. Each cluster is a list of (i, j)
        tuples marking the position of significant points which belong to that
        cluster.
    shape : tuple
        The shape with which to construct the COO matrix.

    Returns
    -------
    scipy.sparse.coo_matrix
        The sparse matrix of significant points.

    Examples
    --------
    >>> from hic3defdr.util.clusters import clusters_to_coo
    >>> coo = clusters_to_coo([[(1, 2), (1, 1)], [(4, 4),  (3, 4)]], (5, 5))
    >>> coo.toarray()
    array([[False, False, False, False, False],
           [False,  True,  True, False, False],
           [False, False, False, False, False],
           [False, False, False, False,  True],
           [False, False, False, False,  True]])
    """
    if not clusters:
        return sparse.coo_matrix(shape, dtype=bool)
    i, j = zip(*[loop for cluster in clusters for loop in cluster])
    return sparse.coo_matrix((np.ones(len(i), dtype=bool), (i, j)), shape=shape)


def clusters_to_pixel_set(clusters):
    """
    Converts a list of clusters to a set of pixels.

    This function has no callers and is usually used as a one-liner.

    Parameters
    ----------
    clusters : list of list of tuple
        The outer list is a list of clusters. Each cluster is a list of (i, j)
        tuples marking the position of significant points which belong to that
        cluster.

    Returns
    -------
    set of tuple
        Each tuple is of the form (i, j) and marks the position of a significant
        point in the clustering.
    """
    return set().union(*clusters)


def cluster_to_slices(cluster, width=21):
    """
    Computes a square row and column slice of a specified width centered on a
    given cluster.

    Parameters
    ----------
    cluster : list of tuple
        A list of (i, j) tuples marking the position of significant points which
        belong to the cluster.
    width : int
        Should be odd. Specifies the side length of the square slice.

    Returns
    -------
    slice, slice
        The row and column slice, respectively.
    """
    w = (width - 1) / 2
    r_idx, c_idx = zip(*cluster)
    r_center = int(np.mean(r_idx))
    c_center = int(np.mean(c_idx))
    return slice(r_center - w, r_center + w + 1), \
        slice(c_center - w, c_center + w + 1)


def filter_clusters_by_distance(clusters, min_dist, max_dist):
    """
    Filters a list of clusters by distance.

    Parameters
    ----------
    clusters : list of list of tuple
        The outer list is a list of clusters. Each cluster is a list of (i, j)
        tuples marking the position of significant points which belong to that
        cluster.
    min_dist, max_dist : int or None
        Specify a range of distances in bin units to filter by (inclusive). If
        either ``min_dist`` or ``max_dist`` is None, the distance bin will be
        considered unbounded on that end.

    Returns
    -------
    list of list of tuple
        The clusters that are within the distance range requested.
    """
    dist_idx = np.ones(len(clusters), dtype=bool)
    dist = np.array([np.mean([j-i for i, j in c]) for c in clusters])
    if min_dist is not None:
        dist_idx[dist < min_dist] = False
    if max_dist is not None:
        dist_idx[dist > max_dist] = False
    return [c for idx, c in zip(dist_idx, clusters) if idx]



import numpy as np
import mpl_scatter_density  # noqa, side-effect on import
import matplotlib.pyplot as plt
from matplotlib.colors import SymLogNorm
from matplotlib.lines import Line2D

from lib5c.util.plotting import plotter


@plotter
def plot_ma(data, sig_idx, loop_idx=None, names=None, s=-1, nonloop_s=None,
            density_dpi=72, vmax=None, nonloop_vmax=None, ax=None, legend=True,
            **kwargs):
    """
    Plots an MA plot.

    Parameters
    ----------
    data : np.ndarray
        Unlogged values to use for computing M and A for each pixel (rows) in
        each condition (columns).
    sig_idx : np.ndarray
        Boolean vector indicating which pixels among those that are in
        ``loop_idx`` are significantly differential.
    loop_idx : np.ndarraym, optional
        Boolean vector indicating which pixels in ``data`` are in loops. Pass
        None if all pixels in ``data`` are in loops.
    names : tuple of str, optional
        The names of the two conitions being compared.
    s : float
        The marker size to pass to `ax.scatter()`. Pass -1 to use
        `ax.scatter_density()` instead, avoiding plotting each point separately.
        See Notes below for more details and caveats.
    nonloop_s : float, optional
        Pass a separate marker size to use specifically for the non-loop pixels
        if `loop_idx` is not None. Useful for drawing just the non-loop pixels
        as a density by passing `s=1, nonloop_s=-1`. Pass None to use `s` as the
        size for both loop and non-loop pixels.
    density_dpi : int
        If `s` is -1 this specifies the DPI to use for the density grid.
    vmax, nonloop_vmax : float, optional
        The vmax to use for `ax.scatter_density()` if `s` or `nonloop_s` is -1,
        respectively. Pass None to choose values automatically.
    ax : pyplot axis
        The axis to plot to. Must have been created with
        `projection='scatter_density'`. Pass None to create a new axis.
    legend : bool
        Pass True to add a legend. Note that passing `legend='outside'` is not
        supported.
    kwargs : kwargs
        Typical plotter kwargs.

    Returns
    -------
    pyplot axis
        The axis plotted on.

    Notes
    -----
    It is recommended to use `ax.scatter_density()` from the
    `mpl_scatter_density` module to plot the data by passing `s=-1`. This avoids
    the massive slowdown experienced when plotting one separate dot for each
    pixel in the genome in the case where `loop_idx=None`. In order to support
    `ax.scatter_density()`, a new pyplot figure and axis must be created with
    `projection='scatter_density'`. Therefore, even though this function is
    decorated with `@plotter`, it has a non-standard behavior for interpreting
    the `ax` kwarg: if `ax=None` is passed, it will create a new figure and axis
    rather than re-using the current axis.
    """
    # compute MA
    m = np.log2(data[:, 0]) - np.log2(data[:, 1])
    a = 0.5 * (np.log2(data[:, 0]) + np.log2(data[:, 1]))

    # resolve ax
    if ax is None:
        if s == -1 or nonloop_s == -1:
            fig = plt.figure()
            ax = fig.add_subplot(1, 1, 1, projection='scatter_density')
        else:
            ax = plt.gca()
    else:
        if (s == -1 or nonloop_s == -1) and not hasattr(ax, 'scatter_density'):
            raise ValueError('ax passed to plot_ma() was not created with '
                             '`projection=\'scatter_density\'`')

    # resolve vmax and create normalization objects
    vmax = len(sig_idx)/1000. if not vmax else vmax
    norm = SymLogNorm(1., vmin=0, vmax=vmax)
    if loop_idx is None:
        nonloop_vmax = None
        nonloop_norm = None
    else:
        nonloop_vmax = (len(loop_idx)-len(sig_idx))/1000. \
            if not nonloop_vmax else nonloop_vmax
        nonloop_norm = SymLogNorm(1., vmin=0, vmax=nonloop_vmax)

    # resolve scatter_fn and scatter_kwargs
    scatter_fn = ax.scatter_density if s == -1 else ax.scatter
    scatter_kwargs = {'dpi': density_dpi, 'norm': norm} if s == -1 else {'s': s}
    nonloop_fn = scatter_fn if nonloop_s is None \
        else ax.scatter_density if nonloop_s == -1 else ax.scatter
    nonloop_kwargs = dict(scatter_kwargs) if nonloop_s is None \
        else {'dpi': density_dpi, 'norm': nonloop_norm} if nonloop_s == -1 \
        else {'s': s}
    if nonloop_s is None and s == -1:
        nonloop_kwargs['norm'] = nonloop_norm

    # apparatus for tracking plot limits
    # we need to do this because ax.scatter_density() resets the axes limits on
    # each call; if s != -1 we will leave the limits alone
    limits = {'xmin': 0, 'xmax': 0, 'ymax': 0}

    def update_lim():
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        limits['xmin'] = min(xmin, limits['xmin'])
        limits['xmax'] = max(xmax, limits['xmax'])
        limits['ymax'] = max(ymax, -ymin, limits['ymax'])

    # prepare colors and labels
    colors = ['gray', 'purple', 'red', 'blue']
    labels = [
        'non-loop pixels',
        'constitutive loop pixels',
        '%s-specific loop pixels' % names[0]
        if names is not None else 'weakening loop pixels',
        '%s-specific loop pixels' % names[1]
        if names is not None else 'strengthening loop pixels'
    ]

    # plot
    include_non_loops = False
    if loop_idx is None:
        loop_idx = np.ones_like(a, dtype=bool)
    else:
        include_non_loops = True
        # non-loop pixels
        nonloop_fn(a[~loop_idx], m[~loop_idx], color=colors[0], label=labels[0],
                   **nonloop_kwargs)
        update_lim()
    # constitutive pixels
    scatter_fn(a[loop_idx][~sig_idx], m[loop_idx][~sig_idx],
               color=colors[1], label=labels[1], **scatter_kwargs)
    update_lim()
    # differential pixels
    if np.any(m[loop_idx][sig_idx] > 0):
        scatter_fn(a[loop_idx][sig_idx][m[loop_idx][sig_idx] > 0],
                   m[loop_idx][sig_idx][m[loop_idx][sig_idx] > 0],
                   color=colors[2], label=labels[2], **scatter_kwargs)
        update_lim()
    if np.any(m[loop_idx][sig_idx] < 0):
        scatter_fn(a[loop_idx][sig_idx][m[loop_idx][sig_idx] < 0],
                   m[loop_idx][sig_idx][m[loop_idx][sig_idx] < 0],
                   color=colors[3], label=labels[3], **scatter_kwargs)
        update_lim()

    # cleanup
    if s == -1 or nonloop_s == -1:
        ax.set_xlim((limits['xmin'], limits['xmax']))
        ax.set_ylim((-limits['ymax'], limits['ymax']))
    ax.set_xlabel('mean log interaction strength')
    ax.set_ylabel('log fold change' +
                  (' (%s over %s)' % tuple(names) if names is not None else ''))

    # add custom legend
    if legend:
        legend_elements = [
            Line2D(
                [0], [0], markersize=10, marker='o', markerfacecolor=color,
                label=label, color=(0, 0, 0, 0), markeredgecolor=(0, 0, 0, 0)
            )
            for color, label in zip(colors, labels)
        ]
        if not include_non_loops:
            legend_elements = legend_elements[1:]
        ax.legend(handles=legend_elements)



import numpy as np
import scipy.stats as stats
from scipy.special import gammaln
from scipy.optimize import newton, brentq

from lib5c.util.mathematics import gmean

from hic3defdr.util.logging import eprint
from hic3defdr.util.progress import tqdm_maybe as tqdm


def logpmf(k, m, phi):
    """
    Log of the PMF of the negative binomial distribution, parameterized by its
    mean ``m`` and dispersion ``phi``. Vectorized.

    Parameters
    ----------
    k : int
        The number of counts observed.
    m : float
        The mean parameter.
    phi : float
        The dispersion parameter.

    Returns
    -------
    float
        The log of the probability of observing ``k`` counts.
    """
    r = 1. / phi
    return gammaln(r+k) - gammaln(k+1) - gammaln(r) + \
        r*np.log(r) - r*np.log(r+m) + k*np.log(m) - k*np.log(r+m)


def mvr(mean, disp):
    """
    Negative binomial fixed-dispersion mean-variance relationship. Vectorized.

    Parameters
    ----------
    mean, disp : float
        The mean and dispersion of a NB distribution, respectively.

    Returns
    -------
    float
        The variance of that NB distribution.
    """
    return mean + mean**2 * disp


def inverse_mvr(mean, var):
    """
    Inverse function of the negative binomial fixed-dispersion mean-variance
    relationship. Vectorized.

    Parameters
    ----------
    mean, var : float
        The mean and variance of a NB distribution, respectively.

    Returns
    -------
    float
        The dispersion of that NB distribution.
    """
    return (var-mean) / mean**2


def fit_mu_hat(x, b, alpha, verbose=True):
    """
    Numerical MLE fitter for the mean parameter of the scaled NB model under
    fixed dispersion. Vectorized.

    See the following colab notebook for background and derivation:
    https://colab.research.google.com/drive/1SgMMvc3XhfIXoBx8tsyJt-yyBDlRFQCJ

    Parameters
    ----------
    x : np.ndarray
        The vector of observed counts.
    b : np.ndarray
        The vector of scaling factors, parallel to ``x``.
    alpha : np.ndarray
        The vector of dispersions, parallel to ``x``.
    verbose : bool
        Pass False to silence reporting of progress to stderr.

    Returns
    -------
    float
        The MLE of the mean parameter.

    Examples
    --------
    >>> import numpy as np
    >>> from hic3defdr.util.scaled_nb import fit_mu_hat

    3 pixels, 2 reps (matrices):
    >>> x = np.array([[1, 2],
    ...               [3, 4],
    ...               [5, 6]])
    >>> b = np.array([[0.9, 1.1],
    ...               [0.8, 1.2],
    ...               [0.7, 1.3]])
    >>> alpha = np.array([[0.1, 0.2],
    ...                   [0.3, 0.4],
    ...                   [0.5, 0.6]])
    >>> fit_mu_hat(x, b, alpha)
    array([1.47251127, 3.53879843, 5.86853465])

    broadcast dispersion down the pixels:
    >>> fit_mu_hat(x, b, np.array([0.1, 0.2]))
    array([1.47251127, 3.53749833, 5.85554075])

    broadcast dispersion across the reps:
    >>> fit_mu_hat(x, b, np.array([0.1, 0.2, 0.3])[:, None])
    array([1.49544092, 3.51679438, 5.73129492])

    1 pixel, two reps (vectors):
    >>> fit_mu_hat(np.array([1, 2]), np.array([0.9, 1.1]), np.array([0.1, 0.2]))
    array([1.47251127])

    broadcast dispersion across reps:
    >>> fit_mu_hat(np.array([1, 2]), np.array([0.9, 1.1]), 0.1)
    array([1.49544092])

    one pixel is fitted with newton, the second is fitted with brentq
    >>> x = np.array([[2, 3, 4, 2],
    ...               [6, 9, 3, 1]])
    >>> b = np.array([[0.45, 0.53, 0.088, 0.091],
    ...               [0.70, 0.83, 0.14,  0.15 ]])
    >>> alpha = np.array([[0.0071, 0.0071, 0.0073, 0.0073],
    ...                   [0.0070, 0.0070, 0.0072, 0.0072]])
    >>> fit_mu_hat(x, b, alpha)
    array([ 9.5900971 , 10.45962955])
    """
    assert np.all((alpha > 0) & np.isfinite(alpha))
    assert np.all((x >= 0) & np.isfinite(x))
    assert np.all((b > 0) & np.isfinite(b))

    def f(mu_hat):
        if hasattr(mu_hat, 'ndim') and mu_hat.ndim < b.ndim and mu_hat.ndim > 0:
            mu_hat = mu_hat[:, None]
        return np.sum((x - mu_hat*b) / (mu_hat + alpha * mu_hat**2 * b),
                      axis=-1)

    if not x.ndim == 2:
        # only one pixel, no need to parallelize
        root = np.array([-1.0])
        failed = np.array([True])
    else:
        # multiple pixels, use newton
        root, converged, zero_der = newton(
            f, np.mean(x / b, axis=1), maxiter=100, full_output=True)
        failed = ~converged | zero_der
        failed[root <= 0] = True  # fail points with negative mu
        failed[root >= np.sqrt(np.finfo(float).max) / 1e10] = True  # overflow
        failed[~np.isclose(f(root), 0, atol=1e-5)] = True  # these aren't close

    if np.any(failed):
        eprint('some points failed, fixing with brentq',
               skip=x.ndim != 2 or not verbose)
        for idx in tqdm(np.where(failed)[0],
                        disable=x.ndim != 2 or not verbose):
            lower = 10 * np.finfo(float).eps
            upper = np.mean(x[idx] / b[idx])
            counter = 0
            while True:
                try:
                    root[idx] = brentq(
                        f if np.isscalar(f(lower)) else lambda y: f(y)[idx],
                        lower, upper)
                    break
                except ValueError:
                    upper *= 2
                    counter += 1
                    if counter > 100:
                        raise ValueError('bracketing interval not found within '
                                         '100 doublings')
    assert np.allclose(f(root), 0, atol=1e-5)
    return root


def equalize(data, f, alpha):
    """
    Given known scaling factors ``f`` and a known dispersion ``alpha``, creates
    common-scale pseudodata from raw values ``data``.

    See https://rdrr.io/bioc/edgeR/src/R/equalizeLibSizes.R

    Parameters
    ----------
    data : np.ndarray
        Matrix of raw data to equalize. Rows are pixels, columns are replicates.
    f : np.ndarray
        Matrix of combined scaling factors for each pixel.
    alpha : float
       Single fixed dispersion to use during equalization.

    Returns
    -------
    np.ndarray
        Matrix of equalized data.
    """
    f_mean = gmean(f, pseudocount=0, axis=1)
    mu_hat = fit_mu_hat(data, f, alpha)
    mu_in = mu_hat[:, None] * f
    mu_out = mu_hat * f_mean
    pseudodata = np.zeros_like(data, dtype=float)
    for i in range(data.shape[1]):
        pseudodata[:, i] = q2qnbinom(data[:, i], mu_in[:, i], mu_out, alpha)
    return pseudodata


def q2qnbinom(x, mu_in, mu_out, alpha):
    """
    Converts values between two NB distributions with different means but the
    same dispersion.

    See https://rdrr.io/bioc/edgeR/src/R/q2qnbinom.R

    Parameters
    ----------
    x : np.ndarray
        Vector of values to convert.
    mu_in, mu_out : np.ndarray
        Vectors of means to convert between.
    alpha : np.ndarray or float
        Single dispserion (to use for all ``x``) or vector of dispersions (one
        per ``x``) to hold constant during conversion.

    Returns
    -------
    np.ndarray
        ``x`` converted from ``mu_in`` to ``mu_out``.
    """
    # force minimum mu of 0.25 for stability
    high_idx = (mu_in >= 0.25) & (mu_out >= 0.25)
    mu_in[~high_idx] = 0.25
    mu_out[~high_idx] = 0.25

    # compute fano factors
    r_in = 1 + alpha * mu_in
    r_out = 1 + alpha * mu_out

    # compute variances
    v_in = mu_in * r_in
    v_out = mu_out * r_out

    # mark points as left or right tail
    right_idx = x >= mu_in

    # construct normal and gamma distributions
    norm_in = stats.norm(mu_in, np.sqrt(v_in))
    norm_out = stats.norm(mu_out, np.sqrt(v_out))
    gamma_in = stats.gamma(mu_in/r_in, scale=r_in)
    gamma_out = stats.gamma(mu_out/r_out, scale=r_out)

    # convert left and right tail separately
    q_norm = np.zeros_like(mu_in)
    q_gamma = np.zeros_like(mu_in)
    q_norm[right_idx] = norm_out.isf(norm_in.sf(x))[right_idx]
    q_norm[~right_idx] = norm_out.ppf(norm_in.cdf(x))[~right_idx]
    q_gamma[right_idx] = gamma_out.isf(gamma_in.sf(x))[right_idx]
    q_gamma[~right_idx] = gamma_out.ppf(gamma_in.cdf(x))[~right_idx]

    # compute pseudocounts as average
    pseudocounts = (q_norm + q_gamma) / 2

    # clip to zero
    pseudocounts[~(pseudocounts >= 0)] = 0

    return pseudocounts



import numpy as np
from scipy.interpolate import interp1d

from lib5c.util.mathematics import gmean

from hic3defdr.util.binning import equal_bin
from hic3defdr.util.progress import tqdm_maybe as tqdm


def no_scaling(data):
    """
    Computes dummy size factors of 1 for each replicate.

    Parameters
    ----------
    data : np.ndarray
        Rows correspond to pixels, columns correspond to replicates.

    Returns
    -------
    np.ndarray
        Vector of size factors, per replicate.
    """
    return np.ones(data.shape[1], dtype=float)


def median_of_ratios(data, filter_zeros=True):
    """
    Computes size factors for a dataset using the median of ratios method.

    Parameters
    ----------
    data : np.ndarray
        Rows correspond to pixels, columns correspond to replicates.
    filter_zeros : bool
        Pass True to filter out pixels with a zero value in any replicate. Pass
        False to include all data.

    Returns
    -------
    np.ndarray
        Vector of size factors, per replicate.
    """
    idx = np.all(data > 0, axis=1) if filter_zeros \
        else np.ones(data.shape[0], dtype=bool)
    return np.median(data[idx, :] /
                     gmean(data[idx, :], axis=1)[:, None], axis=0)


def simple_scaling(data):
    """
    Computes size factors for a dataset using a simple scaling method.

    Parameters
    ----------
    data : np.ndarray
        Rows correspond to pixels, columns correspond to replicates.

    Returns
    -------
    np.ndarray
        Vector of size factors, per replicate.
    """
    s = np.sum(data, axis=0)
    return s / gmean(s)


def conditional(data, dist, fn, n_bins=None):
    """
    Applies a size factor computing function ``fn`` to ``data`` conditioning on
    ``dist``, optionally binning ``dist`` into ``n_bins`` equal-number bins.

    If ``n_bins`` is not None, the final size factors will be interpolated for
    interaction distances falling between the bins to avoid bin edge effects.

    Parameters
    ----------
    data : np.ndarray
        Rows correspond to pixels, columns correspond to replicates.
    dist : np.ndarray
        The distance of each pixel in ``data``.
    fn : function
        A function that computes a size factor given some data.
    n_bins : int, optional
        Pass an int to bin distance into this many equal-number bins.
    """
    result = np.zeros_like(data, dtype=float)
    if n_bins:
        bins = equal_bin(dist, n_bins)
        d_per_bin = []
        s_per_bin = []
        for b in tqdm(np.unique(bins)):
            dist_idx = bins == b
            d_per_bin.append(np.mean(dist[dist_idx]))
            s_per_bin.append(fn(data[dist_idx, :]))
        interp_fns = [interp1d(np.array(d_per_bin), np.array(s_per_bin)[:, i],
                               fill_value='extrapolate', assume_sorted=True)
                      for i in range(data.shape[1])]
        for i in range(data.shape[1]):
            result[:, i] = interp_fns[i](dist)
    else:
        for d in tqdm(np.unique(dist)):
            dist_idx = dist == d
            result[dist_idx, :] = fn(data[dist_idx, :])
    return result


def conditional_mor(data, dist, n_bins=None):
    """
    Computes size factors for a dataset using median of ratios normalization,
    conditioning on distance.

    Parameters
    ----------
    data : np.ndarray
        Rows correspond to pixels, columns correspond to replicates.
    dist : np.ndarray
        The distance of each pixel in ``data``
    n_bins : int, optional
        Pass an int to bin distance into this many equal-number bins.

    Returns
    -------
    np.ndarray
        Matrix of size factors, per pixel (rows) and per replicate (columns).
    """
    return conditional(data, dist, median_of_ratios, n_bins=n_bins)


def conditional_scaling(data, dist, n_bins=None):
    """
    Computes size factors for a dataset using simple scaling normalization,
    conditioning on distance.

    Parameters
    ----------
    data : np.ndarray
        Rows correspond to pixels, columns correspond to replicates.
    dist : np.ndarray
        The distance of each pixel in ``data``.
    n_bins : int, optional
        Pass an int to bin distance into this many equal-number bins.

    Returns
    -------
    np.ndarray
        Matrix of size factors, per pixel (rows) and per replicate (columns).
    """
    return conditional(data, dist, simple_scaling, n_bins=n_bins)



import numpy as np
import scipy.sparse as sparse
import scipy.stats as stats
import scipy.ndimage as ndimage

from lib5c.util.distributions import freeze_distribution

from hic3defdr.util.scaled_nb import mvr
from hic3defdr.util.logging import eprint


def perturb_cluster(matrix, cluster, effect, respect_zeros=True):
    """
    Perturbs a specific cluster in a contact matrix with a given effect.

    Operates in-place.

    Based on a notebook linked here:
    https://colab.research.google.com/drive/1dk9kX57ZtlxQ3jubrKL_q2r8LZnSlVwY

    Parameters
    ----------
    matrix : scipy.sparse.spmatrix
        The contact matrix. Must support slicing.
    cluster : list of tuple of int
        A list of (i, j) tuples marking the position of points which belong to
        the cluster which we want to perturb.
    effect : float
        The effect to apply to the cluster. Values in ``matrix`` under the
        cluster footprint will be shifted by this proportion of their original
        value.
    respect_zeros : bool
        Pass True to preserve the sparsity structure of ``matrix`` if it is
        sparse. Has no effect if ``matrix`` is dense.
    """
    # come up with a rectangle that covers the cluster with a 1px buffer
    rs, cs = map(np.array, zip(*cluster))
    r_min = max(np.min(rs) - 1, 0)
    r_max = min(np.max(rs) + 1, matrix.shape[0] - 1)
    c_min = max(np.min(cs) - 1, 0)
    c_max = min(np.max(cs) + 1, matrix.shape[1] - 1)
    r_size = r_max - r_min + 1
    c_size = c_max - c_min + 1
    r_slice = slice(r_min, r_max + 1)
    c_slice = slice(c_min, c_max + 1)

    # build up the effect footprint
    footprint = np.zeros((r_size, c_size), dtype=float)
    footprint[rs - r_min, cs - c_min] = 1
    struct = ndimage.generate_binary_structure(2, 2)
    footprint += ndimage.binary_dilation(
        footprint, structure=struct)
    footprint /= 2

    # apply the effect footprint
    if isinstance(matrix, sparse.spmatrix) and respect_zeros:
        s = matrix[r_slice, c_slice]
        s_coo = s.tocoo()
        r_read_idx = s_coo.row
        c_read_idx = s_coo.col
        r_write_idx = r_read_idx + r_min
        c_write_idx = c_read_idx + c_min
        new_values = s.toarray() * footprint * effect
        matrix[r_write_idx, c_write_idx] += new_values[r_read_idx, c_read_idx]
    else:
        matrix[r_slice, c_slice] += matrix[r_slice, c_slice].toarray() * \
            footprint * effect


def simulate(row, col, mean, disp_fn, bias, size_factors, clusters, beta=0.5,
             p_diff=0.4, trend='mean', verbose=True):
    """
    Simulates raw contact matrices based on ``mean`` and ``disp_fn`` using
    ``bias`` and ``size_factors`` per simulated replicate and perturbing the
    loops specified in ``clusters`` with an effect size of ``beta`` and
    direction chosen at random for ``p_diff`` fraction of clusters.

    Parameters
    ----------
    row, col : np.ndarray
        Row and column indices identifying the location of pixels in ``mean``.
    mean : np.ndarray
        Vector of mean values for each pixel to use as a base to simulate from.
    disp_fn : function
        Function that returns a dispersion given a mean or distance (as
        specified by ``trend``). Will be used to determine the dispersion
        values to use during simulation.
    bias : np.ndarray
        Rows are bins of the full contact matrix, columns are to-be-simulated
        replicates. Each column represents the bias vector to use for simulating
        that replicate.
    size_factors : np.ndarray
        Vector of size factors to use for simulating for each to-be-simulated
        replicate. To use a different size factor at different distance scales,
        pass a matrix whose rows correspond to distance scales and whose columns
        correspond to replicates.
    clusters : list of list of tuple
        The outer list is a list of clusters which represent the locations of
        loops. Each cluster is a list of (i, j) tuples marking the position of
        pixels which belong to that cluster.
    beta : float
        The effect size of the loop perturbations to use when simulating.
        Perturbed loops will be strengthened or weakened by this fraction of
        their original strength.
    p_diff : float or list of float
        Pass a single float to specify the probability that a loop will be
        perturbed across the simulated conditions. Pass four floats to specify
        the probabilities of all four specific perturbations: up in A, down in
        A, up in B, down in B. The remaining loops will be constitutive.
    trend : 'mean' or 'dist'
        Whether ``disp_fn`` returns the smoothed dispersion as a function of
        mean or of interaction distance.
    verbose : bool
        Pass False to silence reporting of progress to stderr.

    Returns
    -------
    classes : np.ndarray
        Vector of ground-truth class labels used for simulation with '\|S25'
        dtype.
    gen : generator of ``scipy.sparse.csr_matrix``
        Generates the simulated raw contact matrices for each simulated
        replicate, in order.
    """
    eprint('  assigning cluster classes', skip=not verbose)
    p = [1 - p_diff, p_diff/4, p_diff/4, p_diff/4, p_diff/4] \
        if type(p_diff) == float else [1 - sum(p_diff)] + list(p_diff)
    classes = np.random.choice(
        np.array(['constit', 'up A', 'down A', 'up B', 'down B'], dtype='|S7'),
        size=len(clusters), p=p)

    # adjust indexing to make mean nonzero
    nonzero_idx = mean > 0
    row = row[nonzero_idx]
    col = col[nonzero_idx]
    mean = mean[nonzero_idx]
    assert np.all(mean > 0)

    eprint('  perturbing clusters', skip=not verbose)
    mean_a_csr = sparse.coo_matrix(
        (mean, (row, col)), shape=(bias.shape[0], bias.shape[0])).tocsr()
    mean_b_csr = sparse.coo_matrix(
        (mean, (row, col)), shape=(bias.shape[0], bias.shape[0])).tocsr()
    for i, cluster in enumerate(clusters):
        if classes[i] == 'up A':
            perturb_cluster(mean_a_csr, cluster, beta)
        elif classes[i] == 'down A':
            perturb_cluster(mean_a_csr, cluster, -beta)
        elif classes[i] == 'up B':
            perturb_cluster(mean_b_csr, cluster, beta)
        elif classes[i] == 'down B':
            perturb_cluster(mean_b_csr, cluster, -beta)

    # convert to COO
    mean_a_coo = mean_a_csr.tocoo()
    mean_b_coo = mean_b_csr.tocoo()
    del mean_a_csr
    del mean_b_csr

    # confirm index alignment
    assert np.all(mean_a_coo.row == row)
    assert np.all(mean_b_coo.row == row)
    assert np.all(mean_a_coo.col == col)
    assert np.all(mean_b_coo.col == col)
    assert np.all(mean_a_coo.data > 0)
    assert np.all(mean_b_coo.data > 0)

    eprint('  renaming cluster classes', skip=not verbose)
    classes[(classes == 'up A') | (classes == 'down B')] = 'A'
    classes[(classes == 'up B') | (classes == 'down A')] = 'B'

    eprint('  preparing generator', skip=not verbose)
    n_sim = size_factors.shape[-1]
    mean_a = mean_a_coo.data
    mean_b = mean_b_coo.data

    def gen():
        for j, m in zip(range(n_sim), [mean_a]*(n_sim/2) + [mean_b]*(n_sim/2)):
            eprint('  biasing and simulating rep %i/%i' % (j+1, n_sim),
                   skip=not verbose)
            # compute aggregate bias factor
            if len(size_factors.shape) == 1:
                f = bias[row, j] * bias[col, j] * size_factors[j]
            else:
                f = bias[row, j] * bias[col, j] * size_factors[col-row, j]
            assert np.all(f > 0)
            assert np.all((m * f) > 0)

            # compute biased mean
            bm = m * f

            # establish cov
            cov = bm if trend == 'mean' else col - row

            # simulate
            yield sparse.coo_matrix(
                (freeze_distribution(
                    stats.nbinom, bm, mvr(bm, disp_fn(cov))).rvs(),
                 (row, col)), shape=(bias.shape[0], bias.shape[0]))\
                .tocsr()

    return classes, gen()



import os
import sys

import numpy as np
import scipy.sparse as sparse
import pandas as pd

from lib5c.util.system import check_outdir
from hiclite.steps.filter import filter_sparse_rows_count
from hiclite.steps.balance import kr_balance

from hic3defdr import HiC3DeFDR
from hic3defdr.logging import eprint
from hic3defdr.clusters import load_clusters
from hic3defdr.progress import tqdm_maybe as tqdm
from hic3defdr.simulation import simulate


class NoBiasFast3DeFDR(HiC3DeFDR):
    def simulate(self, cond, chrom=None, beta=0.5, p_diff=0.4, trend='mean',
                 outdir='sim'):
        """
        Simulates raw contact matrices based on previously fitted scaled means
        and dispersions in a specific condition.

        This function will only work when the design has exactly two conditions
        and equal numbers of replicates per condition.

        Can only be run after ``estimate_dispersions()`` has been run.

        Parameters
        ----------
        cond : str
            Name of the condition to base the simulation on.
        chrom : str, optional
            Name of the chromosome to simulate. Pass None to simulate all
            chromosomes in series.
        beta : float
            The effect size of the loop perturbations to use when simulating.
            Perturbed loops will be strengthened or weakened by this fraction of
            their original strength.
        p_diff : float
            This fraction of loops will be perturbed across the simulated
            conditions. The remainder will be constitutive.
        trend : 'mean' or 'dist'
            The covariate against which dispersion was fitted when calling
            ``estimate_disp()``. Necessary for correct interpretation of the
            fitted dispersion function as a function of mean or of distance.
        outdir : str
            Path to a directory to store the simulated data to.
        """
        if chrom is None:
            for chrom in self.chroms:
                self.simulate(cond, chrom=chrom, beta=beta, p_diff=p_diff,
                              outdir=outdir)
            return
        eprint('simulating data for chrom %s' % chrom)
        # load everything
        bias = self.load_bias(chrom)
        size_factors = self.load_data('size_factors', chrom)
        row = self.load_data('row', chrom)
        col = self.load_data('col', chrom)
        scaled = self.load_data('scaled', chrom)[:, self.design[cond]]
        disp_fn = self.load_disp_fn(cond, chrom)
        clusters = load_clusters(
            self.loop_patterns[cond].replace('<chrom>', chrom))

        # compute pixel-wise mean of normalized data
        mean = np.mean(scaled, axis=1)

        # book keeping
        check_outdir('%s/' % outdir)
        n_sim = size_factors.shape[-1]
        repnames = sum((['%s%i' % (c, i+1) for i in range(n_sim/2)]
                        for c in ['A', 'B']), [])

        # write design to disk if not present
        design_file = '%s/design.csv' % outdir
        if not os.path.isfile(design_file):
            pd.DataFrame(
                {'A': [1]*(n_sim/2) + [0]*(n_sim/2),
                 'B': [0]*(n_sim/2) + [1]*(n_sim/2)},
                dtype=bool,
                index=repnames
            ).to_csv(design_file)

        # overwrite bias and size factors
        bias = np.ones_like(bias, dtype=float)
        size_factors = np.ones(size_factors.shape[1])

        # rewrite size_factor matrix in terms of distance
        if len(size_factors.shape) == 2:
            eprint('  converting size factors')
            dist = col - row
            n_dists = dist.max() + 1
            new_size_factors = np.zeros((n_dists, size_factors.shape[1]))
            for d in tqdm(range(n_dists)):
                idx = np.argmax(dist == d)
                new_size_factors[d, :] = size_factors[idx, :]
            size_factors = new_size_factors

        # scramble bias and size_factors
        bias = bias[:, (np.arange(n_sim)+1) % n_sim]
        if len(size_factors.shape) == 1:
            size_factors = size_factors[(np.arange(n_sim)+3) % n_sim]
        else:
            size_factors = size_factors[:, (np.arange(n_sim)+3) % n_sim]

        # simulate and save
        classes, sim_iter = simulate(
            row, col, mean, disp_fn, bias, size_factors, clusters, beta=beta,
            p_diff=p_diff, trend=trend)
        np.savetxt('%s/labels_%s.txt' % (outdir, chrom), classes, fmt='%s')
        for rep, csr in zip(repnames, sim_iter):
            sparse.save_npz('%s/%s_%s_raw.npz' % (outdir, rep, chrom), csr)


repnames = ['ES_1', 'ES_3', 'NPC_2', 'NPC_4']
chroms = ['chr19']
base_path = 'c:/lab-data/bonev/'
f = NoBiasFast3DeFDR(
    [base_path + '<rep>/<chrom>_raw.npz'.replace('<rep>', repname)
     for repname in repnames],
    [base_path + '<rep>/<chrom>_kr.bias'.replace('<rep>', repname)
     for repname in repnames],
    chroms, 'design.csv', 'output-test',
    loop_patterns={c: base_path + 'clusters_0.025/%s_<chrom>_clusters.json' % c
                   for c in ['ES', 'NPC']}
)
f.run_to_qvalues()
f.simulate('ES', beta=0.0)

infile_pattern = 'sim/<rep>_<chrom>_raw.npz'
repnames = ['A1', 'A2', 'B1', 'B2']
sim_path = 'sim/'

for repname in repnames:
    for chrom in chroms:
        sys.stderr.write('balancing rep %s chrom %s\n' % (repname, chrom))
        infile = infile_pattern.replace('<rep>', repname)\
            .replace('<chrom>', chrom)
        outfile = infile.replace('_raw.npz', '_kr.bias')
        _, bias, _ = kr_balance(
            filter_sparse_rows_count(sparse.load_npz(infile)), fl=0)
        np.savetxt(outfile, bias)

f_sim = HiC3DeFDR(
    raw_npz_patterns=[sim_path + '<rep>_<chrom>_raw.npz'.replace('<rep>', repname) for repname in repnames],
    bias_patterns=[sim_path + '<rep>_<chrom>_kr.bias'.replace('<rep>', repname) for repname in repnames],
    chroms=chroms,
    design=sim_path + 'design.csv',
    outdir='output-sim',
    loop_patterns={'ES': base_path + 'clusters_0.025/ES_<chrom>_clusters.json'}
)
f_sim.run_to_qvalues(estimator='cml')
f_sim.plot_pvalue_distribution(outfile='pvalues_nobias.png')



import numpy as np

from lib5c.algorithms.correlation import \
    make_pairwise_correlation_matrix_from_counts_matrix
from lib5c.plotters.correlation import plot_correlation_matrix

from hic3defdr.plotting.distance_dependence import plot_dd_curves
from hic3defdr.plotting.histograms import plot_pvalue_histogram
from hic3defdr.plotting.dispersion import plot_mvr, plot_ddr
from hic3defdr.plotting.ma import plot_ma
from hic3defdr.plotting.grid import plot_grid
from hic3defdr.plotting.heatmap import plot_heatmap


class PlottingHiC3DeFDR(object):
    """
    Mixin class containing plotting functions for HiC3DeFDR.
    """
    def plot_dd_curves(self, chrom, log=True, **kwargs):
        """
        Plots the distance dependence curve before and after size factor
        adjustment.

        Parameters
        ----------
        chrom : str
            The name of the chromosome to plot the curve for.
        log : bool
            Whether or not to log the axes of the plot.
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis
            The axis plotted on.
        """
        # load everything
        bias = self.load_bias(chrom)
        row = self.load_data('row', chrom)
        col = self.load_data('col', chrom)
        raw = self.load_data('raw', chrom)
        scaled = self.load_data('scaled', chrom)

        # compute balanced
        balanced = np.zeros_like(raw, dtype=float)
        for r in range(self.design.shape[0]):
            balanced[:, r] = raw[:, r] / (bias[row, r] * bias[col, r])

        return plot_dd_curves(row, col, balanced, scaled,
                              repnames=self.design.index, log=log, **kwargs)

    def plot_dispersion_fit(self, cond, xaxis='dist', yaxis='disp',
                            dist_max=None, scatter_fit=-1, scatter_size=36,
                            distance=None, hexbin=False, logx=False, logy=False,
                            **kwargs):
        """
        Plots a hexbin plot of pixel-wise distance vs either dispersion or
        variance, overlaying the estimated and fitted dispersions.

        Parameters
        ----------
        cond : str
            The name of the chromosome and condition, respectively, to plot the
            fit for.
        xaxis : 'mean' or 'dist'
            What to plot on the x-axis.
        yaxis : 'disp' or 'var'
            What to plot on the y-axis.
        dist_max : int
            If ``xaxis`` is 'dist', the maximum distance to include on the plot
            in bin units. Pass None to use ``self.dist_thresh_max``.
        scatter_fit : int
            Pass a nonzero integer to draw the fitted dispersions passed in
            ``disp`` as a scatterplot of ``scatter_fit`` selected points. Pass
            -1 to plot the fitted dispersions passed in ``disp`` as a curve.
            Pass 0 to omit plotting the dispersion estimates altogether.
        scatter_size : int
            The marker size when plotting scatterplots.
        distance : int, optional
            Pick a specific distance in bin units to plot only interactions at
            that distance.
        hexbin : bool
            Pass False to skip plotting the hexbin plot, leaving only the
            estimated variances or dispersions.
        logx, logy : bool
            Whether or not to log the x- or y-axis, respectively.
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis
            The axis plotted on.
        """
        # short circuit to plot_ddr() if possible
        if xaxis == 'dist' and yaxis == 'disp' and scatter_fit == -1 \
                and distance is None and hexbin is False and logx is False \
                and logy is False:
            return self.plot_ddr(cond, dist_max=dist_max,
                                 scatter_size=scatter_size, **kwargs)

        # resolve max_dist
        if dist_max is None:
            dist_max = self.dist_thresh_max

        # identify cond_idx
        cond_idx = self.design.columns.tolist().index(cond)

        # load everything
        disp_idx, _ = self.load_data('disp_idx', 'all')
        scaled = self.load_data(
            'scaled', 'all', idx=disp_idx)[0][:, self.design[cond]]
        disp = self.load_data('disp', 'all')[0][:, cond_idx]
        try:
            disp_per_dist = self.load_data('disp_per_dist')[:, cond_idx]
            idx = np.isfinite(disp_per_dist)
            disp_per_bin = disp_per_dist[idx]
            dist_per_bin = np.arange(self.dist_thresh_max + 1)[idx]
        except IOError:
            disp_per_bin = None
            dist_per_bin = None
        row, _ = self.load_data('row', 'all', idx=disp_idx)
        col, _ = self.load_data('col', 'all', idx=disp_idx)
        dist = col - row

        # compute mean and sample variance
        mean = np.mean(scaled, axis=1)
        var = np.var(scaled, ddof=1, axis=1)

        # resolve distance
        if distance is not None:
            dist_idx = dist == distance
            mean = mean[dist_idx]
            var = var[dist_idx]
            dist = None
            disp = np.ones(dist_idx.sum()) * disp_per_dist[distance]
            dist_per_bin = None
            disp_per_bin = None
            fit_align_dist = False
        else:
            fit_align_dist = xaxis == 'mean' or yaxis == 'var'

        return plot_mvr(
            pixel_mean=mean,
            pixel_var=var,
            pixel_dist=dist,
            pixel_disp_fit=disp,
            dist_per_bin=dist_per_bin,
            disp_per_bin=disp_per_bin,
            fit_align_dist=fit_align_dist,
            xaxis=xaxis, yaxis=yaxis,
            dist_max=dist_max, mean_min=self.mean_thresh,
            scatter_fit=scatter_fit, scatter_size=scatter_size, hexbin=hexbin,
            logx=logx, logy=logy, **kwargs
        )

    def plot_ddr(self, cond, dist_max=None, scatter_size=36, **kwargs):
        """
        Fast alternative to plot_dispersion_fit() that only supports plotting
        distance versus dispersion, with no hexbin or ``scatter_points``
        support.

        Parameters
        ----------
        cond : str
            The name of the chromosome and condition, respectively, to plot the
            fit for.
        dist_max : int
            If ``xaxis`` is 'dist', the maximum distance to include on the plot
            in bin units. Pass None to use ``self.dist_thresh_max``.
        scatter_size : int
            The marker size when plotting scatterplots.
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis
            The axis plotted on.
        """
        # resolve max_dist
        if dist_max is None:
            dist_max = self.dist_thresh_max

        # identify cond_idx
        cond_idx = self.design.columns.tolist().index(cond)

        # load everything
        disp_per_dist = self.load_data('disp_per_dist')[:, cond_idx]
        idx = np.isfinite(disp_per_dist)
        disp_per_bin = disp_per_dist[idx]
        dist_per_bin = np.arange(self.dist_thresh_max + 1)[idx]
        disp_fn = self.load_disp_fn(cond)

        # plot
        return plot_ddr(dist_per_bin, disp_per_bin, disp_fn,
                        scatter_size=scatter_size, **kwargs)

    def plot_pvalue_distribution(self, idx='disp', **kwargs):
        """
        Plots the p-value distribution across all chromosomes.

        Parameters
        ----------
        idx : {'disp', 'loop'}
            Pass 'disp' to plot p-values for all points for which dispersion was
            estimated. Pass 'loop' to plot p-values for all points which are in
            loops (available only if ``loop_patterns`` was passed to the
            constructor).
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis
            The axis plotted on.
        """
        # load everything
        if idx == 'loop':
            loop_idx, _ = self.load_data('loop_idx', 'all')
            pvalues, _ = self.load_data('pvalues', 'all', idx=loop_idx)
        elif idx == 'disp':
            pvalues, _ = self.load_data('pvalues', 'all')
        else:
            raise ValueError('idx must be loop or disp')

        # plot
        return plot_pvalue_histogram(pvalues, **kwargs)

    def plot_qvalue_distribution(self, **kwargs):
        """
        Plots the q-value distribution across all chromosomes.

        Parameters
        ----------
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis
            The axis plotted on.
        """
        # load everything
        qvalues, _ = self.load_data('qvalues', 'all')

        # plot
        return plot_pvalue_histogram(qvalues, xlabel='qvalue', **kwargs)

    def plot_ma(self, fdr=0.05, conds=None, include_non_loops=True, s=-1,
                nonloop_s=None, density_dpi=72, vmax=None, nonloop_vmax=None,
                ax=None, legend=True, **kwargs):
        """
        Plots an MA plot for a given chromosome.

        Parameters
        ----------
        fdr : float
            The threshold to use for labeling significantly differential loop
            pixels.
        conds : tuple of str, optional
            Pass a tuple of two condition names to compare those two
            conditions. Pass None to compare the first two conditions.
        include_non_loops : bool
            Whether or not to include non-looping pixels in the MA plot.
        s : float
            The marker size to use for the scatterplot, or -1 to use a
            scatter density plot.
        nonloop_s : float, optional
            Pass a separate marker size to use specifically for the non-loop
            pixels if `include_non_loops` is True. Useful for drawing just the
            non-loop pixels as a density by passing `s=1, nonloop_s=-1`. Pass
            None to use `s` as the size for both loop and non-loop pixels.
        density_dpi : int
            If `s` or `nonloop_s` are -1 this specifies the DPI to use for the
            density grid.
        vmax, nonloop_vmax : float, optional
            The vmax to use for `ax.scatter_density()` if `s` or `nonloop_s` is
            -1, respectively. Pass None to choose values automatically.
        ax : pyplot axis
            The axis to plot to. Must have been created with
            `projection='scatter_density'`. Pass None to create a new axis.
        legend : bool
            Pass True to add a legend. Note that passing `legend='outside'` is
            not supported.
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis
            The axis plotted on.
        """
        # resolve conds
        if conds is None:
            conds = self.design.columns.tolist()[:2]
        cond_idx = [self.design.columns.tolist().index(cond) for cond in conds]

        # load data
        disp_idx, _ = self.load_data('disp_idx', 'all')
        loop_idx, _ = self.load_data('loop_idx', 'all')
        scaled, _ = self.load_data('scaled', 'all', idx=disp_idx)
        qvalues, _ = self.load_data('qvalues', 'all')

        # compute mean
        mean = np.dot(scaled, self.design) / np.sum(self.design, axis=0).values
        mean = mean[:, cond_idx]

        # prepare sig_idx
        sig_idx = qvalues < fdr

        # stuff common kwargs into kwargs
        kwargs['names'] = conds
        kwargs['s'] = s
        kwargs['nonloop_s'] = nonloop_s
        kwargs['density_dpi'] = density_dpi
        kwargs['vmax'] = vmax
        kwargs['nonloop_vmax'] = vmax
        kwargs['ax'] = ax
        kwargs['legend'] = legend

        # plot
        if include_non_loops:
            plot_ma(mean, sig_idx, loop_idx=loop_idx, **kwargs)
        else:
            plot_ma(mean[loop_idx], sig_idx, **kwargs)

    def plot_correlation_matrix(self, stage='scaled', idx='loop',
                                correlation='spearman', colorscale=(0.75, 1.0),
                                **kwargs):
        """
        Plots a matrix of pairwise correlations among all replicates.

        Parameters
        ----------
        stage : {'raw', 'scaled'}
            Specify the stage of the data to compute correlations between.
        idx : {'disp', 'loop'}
            Pass 'disp' to compute correlations for all points for which
            dispersion was estimated. Pass 'loop' to compute correlations for
            all points which are in loops (available only if ``loop_patterns``
            was passed to the constructor).
        correlation : {'spearman', 'pearson'}
            Which correlation coefficient to compute.
        colorscale : tuple of float
            The min and max values of the correlation to use to color the
            matrix.
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis
            The axis plotted on.
        """
        # resolve idx
        if idx == 'disp':
            idx, _ = self.load_data('disp_idx', 'all')
        elif idx == 'loop':
            idx = (
                self.load_data('disp_idx', 'all')[0],
                self.load_data('loop_idx', 'all')[0]
            )
        else:
            raise ValueError('idx must be \'disp\' or \'loop\'')

        # load data
        data = self.load_data(stage, 'all', idx=idx)[0].T

        # plot
        return plot_correlation_matrix(
            make_pairwise_correlation_matrix_from_counts_matrix(
                data, correlation=correlation
            ),
            label_values=self.design.index.tolist(),
            colorscale=colorscale,
            **kwargs
        )

    def plot_heatmap(self, rep, chrom, row_slice, col_slice, stage='scaled',
                     cmap='Reds', vmin=0, vmax=100, **kwargs):
        """
        Plots a simple heatmap of a slice of the contact matrix.

        Parameters
        ----------
        rep : str
            The rep to plot.
        chrom : str
            The chromosome to plot.
        row_slice, col_slice : slice
            The row and column slice, respectively, to plot.
        stage : {'raw', 'scaled'}
            The stage of the data to plot.
        cmap : matplotlib colormap
            The colormap to use for the heatmap.
        vmin, vmax : float
            The vmin and vmax to use for the heatmap colorscale.
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis
            The axis plotted on.
        """
        rep_idx = self.design.index.tolist().index(rep)
        row = self.load_data('row', chrom)
        col = self.load_data('col', chrom)
        data = self.load_data(stage, chrom)[:, rep_idx]
        plot_heatmap(row, col, data, row_slice, col_slice, cmap=cmap, vmin=vmin,
                     vmax=vmax, **kwargs)

    def plot_grid(self, chrom, i, j, w, vmax=100, fdr=0.05, cluster_size=3,
                  fdr_vmid=0.05,
                  color_cycle=('blue', 'green', 'purple', 'yellow', 'cyan',
                               'red'),
                  despine=False, **kwargs):
        """
        Plots a combination visualization grid focusing on a specific pixel on a
        specific chromosome, combining heatmaps, cluster outlines, and
        stripplots.

        Parameters
        ----------
        chrom : str
            The name of the chromosome to slice matrices from.
        i, j : int
            The row and column index of the pixel to focus on.
        w : int
            The size of the heatmap will be ``2*w + 1`` bins in each dimension.
        vmax : float
            The maximum of the colorscale to use when plotting normalized
            heatmaps.
        fdr : float
            The FDR threshold to use when outlining clusters.
        cluster_size : int
            The cluster size threshold to use when outlining clusters.
        fdr_vmid : float
            The FDR value at the middle of the colorscale used for plotting the
            q-value heatmap.
        color_cycle : list of matplotlib colors
            The color cycle to use over conditions.
        kwargs : kwargs
            Typical plotter kwargs.

        Returns
        -------
        pyplot axis, grid of pyplot axes, function
            The first pyplot axis returned is injected by ``@plotter``.
            The grid of pyplot axes is the second return value from the call to
            ``plt.subplots()`` that is used to create the grid.
            The function takes two args, an FDR and a cluster size, and redraws
            the cluster outlines using the new parameters.
        """
        # load everything
        row = self.load_data('row', chrom)
        col = self.load_data('col', chrom)
        raw = self.load_data('raw', chrom)
        scaled = self.load_data('scaled', chrom)
        disp_idx = self.load_data('disp_idx', chrom)
        loop_idx = self.load_data('loop_idx', chrom) \
            if self.loop_patterns else np.ones(disp_idx.sum(), dtype=bool)
        mu_hat_alt = self.load_data('mu_hat_alt', chrom)
        mu_hat_null = self.load_data('mu_hat_null', chrom)
        qvalues = self.load_data('qvalues', chrom)

        return plot_grid(i, j, w, row, col, raw, scaled, mu_hat_alt,
                         mu_hat_null, qvalues, disp_idx, loop_idx, self.design,
                         fdr, cluster_size, vmax=vmax, fdr_vmid=fdr_vmid,
                         color_cycle=color_cycle, despine=despine, **kwargs)
